# ** header v3.0
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2021 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see
# <https://www.gnu.org/licenses/>.
#
# ** end header
#
module Entity

export create_entity,
    create_parent, create_property, create_property_entity, create_record, create_recordtype

export get_id, get_role, get_name

export set_id, set_role, set_name

using CaosDB

""" 
Struct containing a pointer to the wrapped cpp entity
object. Meant for internal use; use `CaosDB.Entity.create_entity` to
create an entity. 
"""
mutable struct _Entity
    wrapped_entity::Ptr{Cvoid}

    function _Entity(managed_by_julia::Bool = false)

        entity = new()

        if managed_by_julia
            function f(t)
                ccall(
                    (:caosdb_entity_delete_entity, CaosDB.library_name),
                    Cint,
                    (Ref{_Entity},),
                    Ref{_Entity}(t),
                )
            end
            finalizer(f, entity)
        end

        return entity
    end
end


""" 
Struct containing a pointer to the wrapped cpp parent
object. Meant for internal use; use `CaosDB.Entity.create_parent` to
create an parent. 
"""
mutable struct _Parent
    wrapped_parent::Ptr{Cvoid}

    function _Parent(managed_by_julia::Bool = false)

        parent = new()

        if managed_by_julia
            function f(t)
                ccall(
                    (:caosdb_entity_delete_parent, CaosDB.library_name),
                    Cint,
                    (Ref{_Parent},),
                    Ref{_Parent}(t),
                )
            end
            finalizer(f, parent)
        end

        return parent
    end
end


""" 
Struct containing a pointer to the wrapped cpp property
object. Meant for internal use; use `CaosDB.Entity.create_property` to
create an property. 
"""
mutable struct _Property
    wrapped_property::Ptr{Cvoid}

    function _Property(managed_by_julia::Bool = false)

        property = new()

        if managed_by_julia
            function f(t)
                ccall(
                    (:caosdb_entity_delete_property, CaosDB.library_name),
                    Cint,
                    (Ref{_Property},),
                    Ref{_Property}(t),
                )
            end
            finalizer(f, property)
        end

        return property
    end
end

"""
    function create_entity(name::AbstractString = "")

Return a new entity object. If a `name` was provided, set the name of
the entity correspondingly.
"""
function create_entity(name::AbstractString = "")

    entity = Ref{_Entity}(_Entity(true))

    err_code = ccall(
        (:caosdb_entity_create_entity, CaosDB.library_name),
        Cint,
        (Ref{_Entity},),
        entity,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    if name != ""
        set_name(entity, name)
    end

    return entity

end

"""
    function create_recordtype(name::AbstractString = "")

Return a new entity object with role RecordType. If a `name` was
provided, its name is set accordingly.
"""
function create_recordtype(name::AbstractString = "")

    record_type = create_entity(name)

    set_role(record_type, "RecordType")

    return record_type

end

"""
     function create_record(name::AbstractString = "")

Return a new entity object with role Record. If a `name` was provided,
its name is set accordingly.
"""
function create_record(name::AbstractString = "")

    record = create_entity(name)

    set_role(record, "Record")

    return record

end

"""
    function create_property_entity(;
        name::AbstractString = "",
        datatype::AbstractString = "",
        unit::AbstractString = "",
    )

Return a new entity object with role Record. If `name`, `datatype`, or
`unit` were provided, its name, datatype, or unit are set accordingly.
"""
function create_property_entity(;
    name::AbstractString = "",
    datatype::AbstractString = "",
    unit::AbstractString = "",
)

    property = create_entity(name)

    set_role(property, "Property")

    if datatype != ""
        set_datatype(property, datatype)
    end

    if unit != ""
        set_unit(property, unit)
    end

    return property
end

"""
    function create_property(;
        name::AbstractString = "",
        id::AbstractString = "",
        value::AbstractString = "",
     )

Create a property object that can be appended to another `_Entity`. If
`name`, `id`, or `value` are given, the property's name, id, or value
are set accordingly.

!!! info

    This is not an `_Entity` that could be inserted or updated by its
    own but a `_Property` that is to be appended to another `_Entity`.
"""
function create_property(;
    name::AbstractString = "",
    id::AbstractString = "",
    value::AbstractString = "",
)
    property = Ref{_Property}(_Property(true))

    err_code = ccall(
        (:caosdb_entity_create_property, CaosDB.library_name),
        Cint,
        (Ref{_Property},),
        property,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    if name != ""
        set_name(property, name)
    end

    if id != ""
        set_id(property, id)
    end

    if value
        set_value(property, value)
    end

    return property

end

"""
    function create_parent(; name::AbstractString = "", id::AbstractString = "")

Create a parent object that can be appended to another `_Entity`. If
`name`, `id`, or `value` are given, the parent's name, id, or value
are set accordingly.

!!! info

    This is not an `_Entity` that could be inserted or updated by its
    own but a `_Parent` that is to be appended to another `_Entity`.
"""
function create_parent(; name::AbstractString = "", id::AbstractString = "")

    parent = Ref{_Parent}(_Parent(true))

    err_code = ccall(
        (:caosdb_entity_create_parent, CaosDB.library_name),
        Cint,
        (Ref{_Parent},)parent,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    if name != ""
        set_name(parent, name)
    end

    if id != ""
        set_id(parent, id)
    end

    return parent
end

end
