# ** header v3.0
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2021 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see
# <https://www.gnu.org/licenses/>.
#
# ** end header
#

module Transaction

export create_transaction, add_retrieve_by_id, add_query

using CaosDB

"""
Struct containing a pointer to the wrapped cpp transaction
object. Meant for internal use; call
`CaosDB.Transaction.create_transaction` to create a transaction
object.
"""
mutable struct _Transaction
    wrapped_transaction::Ptr{Cvoid}
    function _Transaction(managed_by_julia::Bool = false)
        trans = new()
        if managed_by_julia
            function f(t)
                ccall(
                    (:caosdb_transaction_delete_transaction, CaosDB.library_name),
                    Cint,
                    (Ref{_Transaction},),
                    Ref{_Transaction}(t),
                )
            end
            finalizer(f, trans)
        end
        return trans
    end
end

"""
    create_transaction(name::AbstractString = "default")

Return a transaction created with the connection of the given name. If
none is given, the defult connection is used.
"""
function create_transaction(name::AbstractString = "default")
    connection = CaosDB.Connection.get_connection(name)

    return create_transaction(connection)
end

"""
    function create_transaction(connection::Ref{CaosDB.Connection._Connection})

Return a transactioncreated with the given connection object.
"""
function create_transaction(connection::Ref{CaosDB.Connection._Connection})

    transaction = Ref{_Transaction}(_Transaction(true))

    err_code = ccall(
        (:caosdb_connection_connection_create_transaction, CaosDB.library_name),
        Cint,
        (Ref{CaosDB.Connection._Connection}, Ref{_Transaction}),
        connection,
        transaction,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return transaction
end

"""
    function add_retrieve_by_id(transaction::Ref{_Transaction}, id::AbstractString)

Add a sub-request to retrieve a single entity by its `id` to the given
`transaction`.

!!! info

    This does not execute the transaction.
"""
function add_retrieve_by_id(transaction::Ref{_Transaction}, id::AbstractString)

    err_code = ccall(
        (:caosdb_transaction_transaction_retrieve_by_id, CaosDB.library_name),
        Cint,
        (Ref{_Transaction}, Cstring),
        transaction,
        id,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function add_retrieve_by_id(
        transaction::Ref{_Transaction},
        ids::Vector{T},
    ) where {T<:AbstractString}

Add a sub-request to retrieve several entities by their `ids` to the
given `transaction`.

!!! info

    This does not execute the transaction.
"""
function add_retrieve_by_id(
    transaction::Ref{_Transaction},
    ids::Vector{T},
) where {T<:AbstractString}

    len = Cint(length(ids))
    err_code = ccall(
        (:caosdb_transaction_transaction_retrieve_by_ids, CaosDB.library_name),
        Cint,
        (Ref{_Transaction}, Ptr{Ptr{Cchar}}, Cint),
        transaction,
        ids,
        len,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end


"""
    function add_query(transaction::Ref{_Transaction}, query::AbstractString)

Add a query sub-request to the given `transaction`.

!!! warning

    Only COUNT queris and FIND queries (and no SELECT queries) are
    currently supported.

!!! info

    This dows not execute the transaction
"""
function add_query(transaction::Ref{_Transaction}, query::AbstractString)

    err_code = ccall(
        (:caosdb_transaction_transaction_query, CaosDB.library_name),
        Cint,
        (Ref{_Transaction}, Cstring),
        transaction,
        query,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

end # Transaction
